@@ -37,7 +37,7 @@
 
 #include "core.h"
 #include "common.h"
-#include "allocator_thread.h"
+#include "rdns.h"
 #include "mutex.h"
 
 extern int tcp_read_time_out;
@@ -188,6 +188,7 @@ static int timed_connect(int sock, const struct sockaddr *addr, socklen_t len) {
 
 
 #define INVALID_INDEX 0xFFFFFFFFU
+#define BUFF_SIZE 1024  // used to read responses from proxies.
 static int tunnel_to(int sock, ip_type ip, unsigned short port, proxy_type pt, char *user, char *pass) {
 	char *dns_name = NULL;
 	char hostnamebuf[MSG_LEN_MAX];
@@ -199,8 +200,8 @@ static int tunnel_to(int sock, ip_type ip, unsigned short port, proxy_type pt, c
 	// the range 224-255.* is reserved, and it won't go outside (unless the app does some other stuff with
 	// the results returned from gethostbyname et al.)
 	// the hardcoded number 224 can now be changed using the config option remote_dns_subnet to i.e. 127
-	if(!ip.is_v6 && ip.addr.v4.octet[0] == remote_dns_subnet) {
-		dns_len = at_get_host_for_ip(ip.addr.v4, hostnamebuf);
+	if(!ip.is_v6 && proxychains_resolver >= DNSLF_RDNS_START && ip.addr.v4.octet[0] == remote_dns_subnet) {
+		dns_len = rdns_get_host_for_ip(ip.addr.v4, hostnamebuf);
 		if(!dns_len) goto err;
 		else dns_name = hostnamebuf;
 	}
@@ -221,6 +222,10 @@ static int tunnel_to(int sock, ip_type ip, unsigned short port, proxy_type pt, c
 	int v6 = ip.is_v6;
 	
 	switch (pt) {
+		case RAW_TYPE: {
+			return SUCCESS;
+		}
+		break;
 		case HTTP_TYPE:{
 			if(!dns_len) {
 				if(!inet_ntop(v6?AF_INET6:AF_INET,ip.addr.v6,ip_buf,sizeof ip_buf)) {
@@ -524,8 +529,8 @@ static int chain_step(int ns, proxy_data * pfrom, proxy_data * pto) {
 
 	PFUNC();
 
-	if(!v6 && pto->ip.addr.v4.octet[0] == remote_dns_subnet) {
-		if(!at_get_host_for_ip(pto->ip.addr.v4, hostname_buf)) goto usenumericip;
+	if(!v6 && proxychains_resolver >= DNSLF_RDNS_START && pto->ip.addr.v4.octet[0] == remote_dns_subnet) {
+		if(!rdns_get_host_for_ip(pto->ip.addr.v4, hostname_buf)) goto usenumericip;
 		else hostname = hostname_buf;
 	} else {
 	usenumericip:
@@ -744,6 +749,102 @@ static void gethostbyname_data_setstring(struct gethostbyname_data* data, char*
 }
 
 extern ip_type4 hostsreader_get_numeric_ip_for_name(const char* name);
+struct hostent* proxy_gethostbyname_old(const char *name)
+{
+	static struct hostent hostent_space;
+	static in_addr_t resolved_addr;
+	static char* resolved_addr_p;
+	static char addr_name[256];
+
+	int pipe_fd[2];
+	char buff[256];
+	in_addr_t addr;
+	pid_t pid;
+	int status, ret;
+	size_t l;
+	struct hostent* hp;
+
+	hostent_space.h_addr_list = &resolved_addr_p;
+	*hostent_space.h_addr_list = (char*)&resolved_addr;
+	resolved_addr = 0;
+
+	if(pc_isnumericipv4(name)) {
+		strcpy(buff, name);
+		goto got_buff;
+	}
+
+	gethostname(buff,sizeof(buff));
+	if(!strcmp(buff,name))
+		goto got_buff;
+
+	memset(buff, 0, sizeof(buff));
+
+	// TODO: this works only once, so cache it  ...
+	// 	 later
+	while ((hp=gethostent()))
+		if (!strcmp(hp->h_name,name))
+			return hp;
+#ifdef HAVE_PIPE2
+	ret = pipe2(pipe_fd, O_CLOEXEC);
+#else
+	ret = pipe(pipe_fd);
+	if(ret == 0) {
+		fcntl(pipe_fd[0], F_SETFD, FD_CLOEXEC);
+		fcntl(pipe_fd[1], F_SETFD, FD_CLOEXEC);
+	}
+#endif
+
+	if(ret)
+		goto err;
+	pid = fork();
+	switch(pid) {
+
+		case 0: // child
+			proxychains_write_log("|DNS-request| %s \n", name);
+			close(pipe_fd[0]);
+			dup2(pipe_fd[1],1);
+			close(pipe_fd[1]);
+
+		//	putenv("LD_PRELOAD=");
+			execlp("proxyresolv","proxyresolv",name,NULL);
+			perror("can't exec proxyresolv");
+			exit(2);
+
+		case -1: //error
+			close(pipe_fd[0]);
+			close(pipe_fd[1]);
+			perror("can't fork");
+			goto err;
+
+		default:
+			close(pipe_fd[1]);
+			waitpid(pid, &status, 0);
+			buff[0] = 0;
+			read(pipe_fd[0],&buff,sizeof(buff));
+			close(pipe_fd[0]);
+got_buff:
+			l = strlen(buff);
+			if(l && buff[l-1] == '\n') buff[l-1] = 0;
+			addr = inet_addr(buff);
+			if (addr == (in_addr_t) (-1))
+				goto err_dns;
+			memcpy(*(hostent_space.h_addr_list),
+						&addr ,sizeof(struct in_addr));
+			hostent_space.h_name = addr_name;
+			snprintf(addr_name, sizeof addr_name, "%s", buff);
+			hostent_space.h_length = sizeof (in_addr_t);
+			hostent_space.h_addrtype = AF_INET;
+	}
+	proxychains_write_log("|DNS-response| %s is %s\n",
+			name, inet_ntoa(*(struct in_addr*)&addr));
+	return &hostent_space;
+err_dns:
+	proxychains_write_log("|DNS-response|: %s does not exist\n", name);
+	perror("err_dns");
+err:
+	return NULL;
+}
+
 struct hostent *proxy_gethostbyname(const char *name, struct gethostbyname_data* data) {
 	PFUNC();
 	char buff[256];
@@ -759,24 +860,29 @@ struct hostent *proxy_gethostbyname(const char *name, struct gethostbyname_data*
 	data->hostent_space.h_addrtype = AF_INET;
 	data->hostent_space.h_length = sizeof(in_addr_t);
 
+	if(pc_isnumericipv4(name)) {
+		data->resolved_addr = inet_addr(name);
+		goto retname;
+	}
+
 	gethostname(buff, sizeof(buff));
 
 	if(!strcmp(buff, name)) {
 		data->resolved_addr = inet_addr(buff);
 		if(data->resolved_addr == (in_addr_t) (-1))
-			data->resolved_addr = (in_addr_t) (ip_type_localhost.addr.v4.as_int);
+			data->resolved_addr = (in_addr_t) (IPT4_LOCALHOST.as_int);
 		goto retname;
 	}
 
 	// this iterates over the "known hosts" db, usually /etc/hosts
 	ip_type4 hdb_res = hostsreader_get_numeric_ip_for_name(name);
-	if(hdb_res.as_int != ip_type_invalid.addr.v4.as_int) {
+	if(hdb_res.as_int != IPT4_INVALID.as_int) {
 		data->resolved_addr = hdb_res.as_int;
 		goto retname;
 	}
 	
-	data->resolved_addr = at_get_ip_for_host((char*) name, strlen(name)).as_int;
-	if(data->resolved_addr == (in_addr_t) ip_type_invalid.addr.v4.as_int) return NULL;
+	data->resolved_addr = rdns_get_ip_for_host((char*) name, strlen(name)).as_int;
+	if(data->resolved_addr == (in_addr_t) IPT4_INVALID.as_int) return NULL;
 
 	retname:
 
@@ -789,7 +895,7 @@ struct hostent *proxy_gethostbyname(const char *name, struct gethostbyname_data*
 
 struct addrinfo_data {
 	struct addrinfo addrinfo_space;
-	struct sockaddr sockaddr_space;
+	struct sockaddr_storage sockaddr_space;
 	char addr_name[256];
 };
 
@@ -822,6 +928,33 @@ static int mygetservbyname_r(const char* name, const char* proto, struct servent
 	return ret;
 }
 
+static int looks_like_numeric_ipv6(const char *node)
+{
+	if(!strchr(node, ':')) return 0;
+	const char* p= node;
+	while(1) switch(*(p++)) {
+		case 0: return 1;
+		case ':': case '.':
+		case '0': case '1': case '2': case '3': case '4':
+		case '5': case '6': case '7': case '8': case '9':
+		case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
+		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
+			break;
+		default: return 0;
+	}
+}
+
+static int my_inet_aton(const char *node, struct addrinfo_data* space)
+{
+	int ret;
+	((struct sockaddr_in *) &space->sockaddr_space)->sin_family = AF_INET;
+	ret = inet_aton(node, &((struct sockaddr_in *) &space->sockaddr_space)->sin_addr);
+	if(ret || !looks_like_numeric_ipv6(node)) return ret;
+	ret = inet_pton(AF_INET6, node, &((struct sockaddr_in6 *) &space->sockaddr_space)->sin6_addr);
+	if(ret) ((struct sockaddr_in6 *) &space->sockaddr_space)->sin6_family = AF_INET6;
+	return ret;
+}
+
 int proxy_getaddrinfo(const char *node, const char *service, const struct addrinfo *hints, struct addrinfo **res) {
 	struct gethostbyname_data ghdata;
 	struct addrinfo_data *space;
@@ -830,42 +963,56 @@ int proxy_getaddrinfo(const char *node, const char *service, const struct addrin
 	struct servent se_buf;
 	struct addrinfo *p;
 	char buf[1024];
-	int port;
-	PFUNC();
+	int port, af = AF_INET;
+
+	PDEBUG("proxy_getaddrinfo node:%s service: %s, flags: %d\n",
+		node?node:"",service?service:"",hints?(int)hints->ai_flags:0);
 
-//      printf("proxy_getaddrinfo node %s service %s\n",node,service);
 	space = calloc(1, sizeof(struct addrinfo_data));
 	if(!space) goto err1;
 
-	if(node && !inet_aton(node, &((struct sockaddr_in *) &space->sockaddr_space)->sin_addr)) {
+	if(node && !my_inet_aton(node, space)) {
 		/* some folks (nmap) use getaddrinfo() with AI_NUMERICHOST to check whether a string
 		   containing a numeric ip was passed. we must return failure in that case. */
 		if(hints && (hints->ai_flags & AI_NUMERICHOST)) {
 			free(space);
 			return EAI_NONAME;
 		}
-		hp = proxy_gethostbyname(node, &ghdata);
+		if(proxychains_resolver == DNSLF_FORKEXEC)
+			hp = proxy_gethostbyname_old(node);
+		else
+			hp = proxy_gethostbyname(node, &ghdata);
+
 		if(hp)
 			memcpy(&((struct sockaddr_in *) &space->sockaddr_space)->sin_addr,
 			       *(hp->h_addr_list), sizeof(in_addr_t));
 		else
 			goto err2;
+	} else if(node) {
+		af = ((struct sockaddr_in *) &space->sockaddr_space)->sin_family;
+	} else if(!node && !(hints->ai_flags & AI_PASSIVE)) {
+		af = ((struct sockaddr_in *) &space->sockaddr_space)->sin_family = AF_INET;
+		memcpy(&((struct sockaddr_in *) &space->sockaddr_space)->sin_addr,
+		       (char[]){127,0,0,1}, 4);
 	}
 	if(service) mygetservbyname_r(service, NULL, &se_buf, buf, sizeof(buf), &se);
 
 	port = se ? se->s_port : htons(atoi(service ? service : "0"));
-	((struct sockaddr_in *) &space->sockaddr_space)->sin_port = port;
+	if(af == AF_INET)
+		((struct sockaddr_in *) &space->sockaddr_space)->sin_port = port;
+	else
+		((struct sockaddr_in6 *) &space->sockaddr_space)->sin6_port = port;
 
 	*res = p = &space->addrinfo_space;
 	assert((size_t)p == (size_t) space);
 
-	p->ai_addr = &space->sockaddr_space;
+	p->ai_addr = (void*) &space->sockaddr_space;
 	if(node)
 		snprintf(space->addr_name, sizeof(space->addr_name), "%s", node);
 	p->ai_canonname = space->addr_name;
 	p->ai_next = NULL;
-	p->ai_family = space->sockaddr_space.sa_family = AF_INET;
-	p->ai_addrlen = sizeof(space->sockaddr_space);
+	p->ai_family = space->sockaddr_space.ss_family = af;
+	p->ai_addrlen = af == AF_INET ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);
 
 	if(hints) {
 		p->ai_socktype = hints->ai_socktype;
