@@ -22,6 +22,7 @@
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
+#include <ctype.h>
 #include <errno.h>
 #include <assert.h>
 #include <netdb.h>
@@ -37,7 +38,9 @@
 
 #include "core.h"
 #include "common.h"
+#include "rdns.h"
 
+#undef 		satosin
 #define     satosin(x)      ((struct sockaddr_in *) &(x))
 #define     SOCKADDR(x)     (satosin(x)->sin_addr.s_addr)
 #define     SOCKADDR_2(x)     (satosin(x)->sin_addr)
@@ -69,9 +72,11 @@ unsigned int proxychains_proxy_offset = 0;
 int proxychains_got_chain_data = 0;
 unsigned int proxychains_max_chain = 1;
 int proxychains_quiet_mode = 0;
-int proxychains_resolver = 0;
+enum dns_lookup_flavor proxychains_resolver = DNSLF_LIBC;
 localaddr_arg localnet_addr[MAX_LOCALNET];
 size_t num_localnet_addr = 0;
+dnat_arg dnats[MAX_DNAT];
+size_t num_dnats = 0;
 unsigned int remote_dns_subnet = 224;
 
 pthread_once_t init_once = PTHREAD_ONCE_INIT;
@@ -113,19 +118,28 @@ static void setup_hooks(void) {
 	SETUP_SYM(freeaddrinfo);
 	SETUP_SYM(gethostbyaddr);
 	SETUP_SYM(getnameinfo);
-	SETUP_SYM(close);
 #ifdef IS_SOLARIS
 	SETUP_SYM(__xnet_connect);
 #endif
+	SETUP_SYM(close);
 }
 
 static int close_fds[16];
 static int close_fds_cnt = 0;
 
+static unsigned get_rand_seed(void) {
+#ifdef HAVE_CLOCK_GETTIME
+	struct timespec now;
+	clock_gettime(CLOCK_REALTIME, &now);
+	return now.tv_sec ^ now.tv_nsec;
+#else
+	return time(NULL);
+#endif
+}
+
 static void do_init(void) {
-	srand(time(NULL));
+	srand(get_rand_seed());
 	core_initialize();
-	at_init();
 
 	/* read the config file */
 	get_chain_data(proxychains_pd, &proxychains_proxy_count, &proxychains_ct);
@@ -136,8 +150,9 @@ static void do_init(void) {
 	setup_hooks();
 
 	while(close_fds_cnt) true_close(close_fds[--close_fds_cnt]);
-
 	init_l = 1;
+
+	rdns_init(proxychains_resolver);
 }
 
 static void init_lib_wrapper(const char* caller) {
@@ -208,7 +223,7 @@ static int proxy_from_string(const char *proxystring,
 	   proxystring[next_token++] != ':' ||
 	   proxystring[next_token++] != '/' ||
 	   proxystring[next_token++] != '/') goto inv_string;
-	const char *at = strchr(proxystring+next_token, '@');
+	const char *at = strrchr(proxystring+next_token, '@');
 	if(at) {
 		if(proxytype == RS_PT_SOCKS4)
 			return 0;
@@ -261,24 +276,30 @@ static const char* bool_str(int bool_val) {
 	return "false";
 }
 
+#define STR_STARTSWITH(P, LIT) (!strncmp(P, LIT, sizeof(LIT)-1))
 /* get configuration from config file */
 static void get_chain_data(proxy_data * pd, unsigned int *proxy_count, chain_type * ct) {
 	int count = 0, port_n = 0, list = 0;
-	char buff[1024], type[1024], host[1024], user[1024];
-	char *env;
+	char buf[1024], type[1024], host[1024], user[1024];
+	char *buff, *env, *p;
 	char local_in_addr_port[32];
 	char local_in_addr[32], local_in_port[32], local_netmask[32];
+	char dnat_orig_addr_port[32], dnat_new_addr_port[32];
+	char dnat_orig_addr[32], dnat_orig_port[32], dnat_new_addr[32], dnat_new_port[32];
+	char rdnsd_addr[32], rdnsd_port[8];
 	FILE *file = NULL;
 
 	if(proxychains_got_chain_data)
 		return;
 
+	PFUNC();
+
 	//Some defaults
 	tcp_read_time_out = 4 * 1000;
 	tcp_connect_time_out = 10 * 1000;
 	*ct = DYNAMIC_TYPE;
 
-	env = get_config_path(getenv(PROXYCHAINS_CONF_FILE_ENV_VAR), buff, sizeof(buff));
+	env = get_config_path(getenv(PROXYCHAINS_CONF_FILE_ENV_VAR), buf, sizeof(buf));
 	if( ( file = fopen(env, "r") ) == NULL )
 	{
 	        perror("couldnt read configuration file");
@@ -289,8 +310,17 @@ static void get_chain_data(proxy_data * pd, unsigned int *proxy_count, chain_typ
 	if(env && *env == '1')
 		proxychains_quiet_mode = 1;
 
-	while(fgets(buff, sizeof(buff), file)) {
-		if(buff[0] != '\n' && buff[strspn(buff, " ")] != '#') {
+	while(fgets(buf, sizeof(buf), file)) {
+		buff = buf;
+		/* remove leading whitespace */
+		while(isspace(*buff)) buff++;
+		/* remove trailing '\n' */
+		if((p = strrchr(buff, '\n'))) *p = 0;
+		p = buff + strlen(buff)-1;
+		/* remove trailing whitespace */
+		while(p >= buff && isspace(*p)) *(p--) = 0;
+		if(!*buff || *buff == '#') continue; /* skip empty lines and comments */
+		if(1) {
 			/* proxylist has to come last */
 			if(list) {
 				if(count >= MAX_CHAIN)
@@ -315,24 +345,27 @@ static void get_chain_data(proxy_data * pd, unsigned int *proxy_count, chain_typ
 				pd[count].port = htons((unsigned short) port_n);
 				ip_type* host_ip = &pd[count].ip;
 				if(1 != inet_pton(host_ip->is_v6 ? AF_INET6 : AF_INET, host, host_ip->addr.v6)) {
-					if(*ct == STRICT_TYPE && proxychains_resolver && count > 0) {
+					if(*ct == STRICT_TYPE && proxychains_resolver >= DNSLF_RDNS_START && count > 0) {
 						/* we can allow dns hostnames for all but the first proxy in the list if chaintype is strict, as remote lookup can be done */
+						rdns_init(proxychains_resolver);
 						ip_type4 internal_ip = at_get_ip_for_host(host, strlen(host));
 						pd[count].ip.is_v6 = 0;
 						host_ip->addr.v4 = internal_ip;
-						if(internal_ip.as_int == ip_type_invalid.addr.v4.as_int)
+						if(internal_ip.as_int == IPT4_INVALID.as_int)
 							goto inv_host;
 					} else {
 inv_host:
 						fprintf(stderr, "proxy %s has invalid value or is not numeric\n", host);
 						fprintf(stderr, "non-numeric ips are only allowed under the following circumstances:\n");
-						fprintf(stderr, "chaintype == strict (%s), proxy is not first in list (%s), proxy_dns active (%s)\n\n", bool_str(*ct == STRICT_TYPE), bool_str(count > 0), bool_str(proxychains_resolver));
+						fprintf(stderr, "chaintype == strict (%s), proxy is not first in list (%s), proxy_dns active (%s)\n\n", bool_str(*ct == STRICT_TYPE), bool_str(count > 0), rdns_resolver_string(proxychains_resolver));
 						exit(1);
 					}
 				}
 
 				if(!strcmp(type, "http")) {
 					pd[count].pt = HTTP_TYPE;
+				} else if(!strcmp(type, "raw")) {
+					pd[count].pt = RAW_TYPE;
 				} else if(!strcmp(type, "socks4")) {
 					pd[count].pt = SOCKS4_TYPE;
 				} else if(!strcmp(type, "socks5")) {
@@ -343,28 +376,28 @@ inv_host:
 				if(port_n)
 					count++;
 			} else {
-				if(strstr(buff, "[ProxyList]")) {
+				if(!strcmp(buff, "[ProxyList]")) {
 					list = 1;
-				} else if(strstr(buff, "random_chain")) {
+				} else if(!strcmp(buff, "random_chain")) {
 					*ct = RANDOM_TYPE;
-				} else if(strstr(buff, "strict_chain")) {
+				} else if(!strcmp(buff, "strict_chain")) {
 					*ct = STRICT_TYPE;
-				} else if(strstr(buff, "dynamic_chain")) {
+				} else if(!strcmp(buff, "dynamic_chain")) {
 					*ct = DYNAMIC_TYPE;
-				} else if(strstr(buff, "round_robin_chain")) {
+				} else if(!strcmp(buff, "round_robin_chain")) {
 					*ct = ROUND_ROBIN_TYPE;
-				} else if(strstr(buff, "tcp_read_time_out")) {
+				} else if(STR_STARTSWITH(buff, "tcp_read_time_out")) {
 					sscanf(buff, "%s %d", user, &tcp_read_time_out);
-				} else if(strstr(buff, "tcp_connect_time_out")) {
+				} else if(STR_STARTSWITH(buff, "tcp_connect_time_out")) {
 					sscanf(buff, "%s %d", user, &tcp_connect_time_out);
-				} else if(strstr(buff, "remote_dns_subnet")) {
+				} else if(STR_STARTSWITH(buff, "remote_dns_subnet")) {
 					sscanf(buff, "%s %u", user, &remote_dns_subnet);
 					if(remote_dns_subnet >= 256) {
 						fprintf(stderr,
 							"remote_dns_subnet: invalid value. requires a number between 0 and 255.\n");
 						exit(1);
 					}
-				} else if(strstr(buff, "localnet")) {
+				} else if(STR_STARTSWITH(buff, "localnet")) {
 					if(sscanf(buff, "%s %21[^/]/%15s", user, local_in_addr_port, local_netmask) < 3) {
 						fprintf(stderr, "localnet format error");
 						exit(1);
@@ -405,7 +438,7 @@ inv_host:
 					} else {
 						fprintf(stderr, "# of localnet exceed %d.\n", MAX_LOCALNET);
 					}
-				} else if(strstr(buff, "chain_len")) {
+				} else if(STR_STARTSWITH(buff, "chain_len")) {
 					char *pc;
 					int len;
 					pc = strchr(buff, '=');
@@ -415,10 +448,77 @@ inv_host:
 					}
 					len = atoi(++pc);
 					proxychains_max_chain = (len ? len : 1);
-				} else if(strstr(buff, "quiet_mode")) {
+				} else if(!strcmp(buff, "quiet_mode")) {
 					proxychains_quiet_mode = 1;
-				} else if(strstr(buff, "proxy_dns")) {
-					proxychains_resolver = 1;
+				} else if(!strcmp(buff, "proxy_dns_old")) {
+					proxychains_resolver = DNSLF_FORKEXEC;
+				} else if(!strcmp(buff, "proxy_dns")) {
+					proxychains_resolver = DNSLF_RDNS_THREAD;
+				} else if(STR_STARTSWITH(buff, "proxy_dns_daemon")) {
+					struct sockaddr_in rdns_server_buffer;
+
+					if(sscanf(buff, "%s %15[^:]:%5s", user, rdnsd_addr, rdnsd_port) < 3) {
+						fprintf(stderr, "proxy_dns_daemon format error\n");
+						exit(1);
+					}
+					rdns_server_buffer.sin_family = AF_INET;
+					int error = inet_pton(AF_INET, rdnsd_addr, &rdns_server_buffer.sin_addr);
+					if(error <= 0) {
+						fprintf(stderr, "bogus proxy_dns_daemon address\n");
+						exit(1);
+					}
+					rdns_server_buffer.sin_port = htons(atoi(rdnsd_port));
+					proxychains_resolver = DNSLF_RDNS_DAEMON;
+					rdns_set_daemon(&rdns_server_buffer);
+				} else if(STR_STARTSWITH(buff, "dnat")) {
+					if(sscanf(buff, "%s %21[^ ] %21s\n", user, dnat_orig_addr_port, dnat_new_addr_port) < 3) {
+						fprintf(stderr, "dnat format error");
+						exit(1);
+					}
+					/* clean previously used buffer */
+					memset(dnat_orig_port, 0, sizeof(dnat_orig_port) / sizeof(dnat_orig_port[0]));
+					memset(dnat_new_port, 0, sizeof(dnat_new_port) / sizeof(dnat_new_port[0]));
+
+					(void)sscanf(dnat_orig_addr_port, "%15[^:]:%5s", dnat_orig_addr, dnat_orig_port);
+					(void)sscanf(dnat_new_addr_port, "%15[^:]:%5s", dnat_new_addr, dnat_new_port);
+
+					if(num_dnats < MAX_DNAT) {
+						int error;
+						error =
+						    inet_pton(AF_INET, dnat_orig_addr,
+							      &dnats[num_dnats].orig_dst);
+						if(error <= 0) {
+							fprintf(stderr, "dnat original destination address error\n");
+							exit(1);
+						}
+
+						error =
+						    inet_pton(AF_INET, dnat_new_addr,
+							      &dnats[num_dnats].new_dst);
+						if(error <= 0) {
+							fprintf(stderr, "dnat effective destination address error\n");
+							exit(1);
+						}
+
+						if(dnat_orig_port[0]) {
+							dnats[num_dnats].orig_port =
+							    (short) atoi(dnat_orig_port);
+						} else {
+							dnats[num_dnats].orig_port = 0;
+						}
+
+						if(dnat_new_port[0]) {
+							dnats[num_dnats].new_port =
+							    (short) atoi(dnat_new_port);
+						} else {
+							dnats[num_dnats].new_port = 0;
+						}
+
+						PDEBUG("added dnat: orig-dst=%s orig-port=%d new-dst=%s new-port=%d\n", dnat_orig_addr, dnats[num_dnats].orig_port, dnat_new_addr, dnats[num_dnats].new_port);
+						++num_dnats;
+					} else {
+						fprintf(stderr, "# of dnat exceed %d.\n", MAX_DNAT);
+					}
 				}
 			}
 		}
@@ -432,7 +532,7 @@ inv_host:
 	}
 	*proxy_count = count;
 	proxychains_got_chain_data = 1;
-	PDEBUG("proxy_dns: %s\n", proxychains_resolver ? "ON" : "OFF");
+	PDEBUG("proxy_dns: %s\n", rdns_resolver_string(proxychains_resolver));
 }
 
 /*******  HOOK FUNCTIONS  *******/
@@ -444,6 +544,8 @@ int close(int fd) {
 		errno = 0;
 		return 0;
 	}
+	if(proxychains_resolver != DNSLF_RDNS_THREAD) return true_close(fd);
+
 	/* prevent rude programs (like ssh) from closing our pipes */
 	if(fd != req_pipefd[0]  && fd != req_pipefd[1] &&
 	   fd != resp_pipefd[0] && fd != resp_pipefd[1]) {
@@ -467,6 +569,7 @@ int connect(int sock, const struct sockaddr *addr, unsigned int len) {
 
 	struct in_addr *p_addr_in;
 	struct in6_addr *p_addr_in6;
+	dnat_arg *dnat = NULL;
 	unsigned short port;
 	size_t i;
 	int remote_dns_connect = 0;
@@ -501,6 +604,23 @@ int connect(int sock, const struct sockaddr *addr, unsigned int len) {
 	// check if connect called from proxydns
         remote_dns_connect = !v6 && (ntohl(p_addr_in->s_addr) >> 24 == remote_dns_subnet);
 
+	// more specific first
+	if (!v6) for(i = 0; i < num_dnats && !remote_dns_connect && !dnat; i++)
+		if(dnats[i].orig_dst.s_addr == p_addr_in->s_addr)
+			if(dnats[i].orig_port && (dnats[i].orig_port == port))
+				dnat = &dnats[i];
+
+	if (!v6) for(i = 0; i < num_dnats && !remote_dns_connect && !dnat; i++)
+		if(dnats[i].orig_dst.s_addr == p_addr_in->s_addr)
+			if(!dnats[i].orig_port)
+				dnat = &dnats[i];
+
+	if (dnat) {
+		p_addr_in = &dnat->new_dst;
+		if (dnat->new_port)
+			port = dnat->new_port;
+	}
+
 	if (!v6) for(i = 0; i < num_localnet_addr && !remote_dns_connect; i++) {
 		if((localnet_addr[i].in_addr.s_addr & localnet_addr[i].netmask.s_addr)
 		   == (p_addr_in->s_addr & localnet_addr[i].netmask.s_addr)) {
@@ -539,10 +659,12 @@ struct hostent *gethostbyname(const char *name) {
 	INIT();
 	PDEBUG("gethostbyname: %s\n", name);
 
-	if(proxychains_resolver)
-		return proxy_gethostbyname(name, &ghbndata);
-	else
+	if(proxychains_resolver == DNSLF_FORKEXEC)
+		return proxy_gethostbyname_old(name);
+	else if(proxychains_resolver == DNSLF_LIBC)
 		return true_gethostbyname(name);
+	else
+		return proxy_gethostbyname(name, &ghbndata);
 
 	return NULL;
 }
@@ -551,7 +673,7 @@ int getaddrinfo(const char *node, const char *service, const struct addrinfo *hi
 	INIT();
 	PDEBUG("getaddrinfo: %s %s\n", node ? node : "null", service ? service : "null");
 
-	if(proxychains_resolver)
+	if(proxychains_resolver != DNSLF_LIBC)
 		return proxy_getaddrinfo(node, service, hints, res);
 	else
 		return true_getaddrinfo(node, service, hints, res);
@@ -561,7 +683,7 @@ void freeaddrinfo(struct addrinfo *res) {
 	INIT();
 	PDEBUG("freeaddrinfo %p \n", (void *) res);
 
-	if(!proxychains_resolver)
+	if(proxychains_resolver == DNSLF_LIBC)
 		true_freeaddrinfo(res);
 	else
 		proxy_freeaddrinfo(res);
@@ -574,7 +696,7 @@ int pc_getnameinfo(const struct sockaddr *sa, socklen_t salen,
 	INIT();
 	PFUNC();
 
-	if(!proxychains_resolver) {
+	if(proxychains_resolver == DNSLF_LIBC) {
 		return true_getnameinfo(sa, salen, host, hostlen, serv, servlen, flags);
 	} else {
 		if(!salen || !(SOCKFAMILY(*sa) == AF_INET || SOCKFAMILY(*sa) == AF_INET6))
@@ -621,7 +743,7 @@ struct hostent *gethostbyaddr(const void *addr, socklen_t len, int type) {
 	static char *aliases[1];
 	static struct hostent he;
 
-	if(!proxychains_resolver)
+	if(proxychains_resolver == DNSLF_LIBC)
 		return true_gethostbyaddr(addr, len, type);
 	else {
 
