@@ -7,10 +7,7 @@
  *                                                                         *
  ***************************************************************************/
 
-#undef _POSIX_C_SOURCE
-#define _POSIX_C_SOURCE 200112L
-#undef _XOPEN_SOURCE
-#define _XOPEN_SOURCE 600
+#define _DEFAULT_SOURCE
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -19,6 +16,11 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 
+#ifdef IS_MAC
+#define _DARWIN_C_SOURCE
+#endif
+#include <dlfcn.h>
+
 #include "common.h"
 
 static int usage(char **argv) {
@@ -49,7 +51,7 @@ static void set_own_dir(const char *argv0) {
 	size_t l = strlen(argv0);
 	while(l && argv0[l - 1] != '/')
 		l--;
-	if(l == 0)
+	if(l == 0 || l >= sizeof(own_dir))
 #ifdef SUPER_SECURE
 		memcpy(own_dir, "/dev/null/", 11);
 #else
@@ -111,7 +113,9 @@ int main(int argc, char *argv[]) {
 
 	// search DLL
 
-	set_own_dir(argv[0]);
+	Dl_info dli;
+	dladdr(own_dir, &dli);
+	set_own_dir(dli.dli_fname);
 
 	i = 0;
 
